TEXT_TO_SPEECH_CHUNK_SIZE_SECONDS = 0.1
PER_CHUNK_ALLOWANCE_SECONDS = 0.01
ALLOWED_IDLE_TIME = 15
SENTENCE_ENDINGS = [".", "!", "?", "\n"]
CHECK_HUMAN_PRESENT_MESSAGE_CHOICES = [
    "Hello?",
    "Are you there?",
    "Are you still there?",
    "Hi, are you there?",
]
